
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "csi.h"

#ifdef	__MSVCRT__
#define LL_FORMAT	"I64"
#else
#define LL_FORMAT	"ll"
#endif

extern void	Error(),ReadSliceImage();

#ifndef	RECURSION_LEVEL
#define RECURSION_LEVEL	6
#endif
#define TOTAL_LEVEL	24

static double	xx,yy,zz,xy2,yz2,zx2,L0_Lc;
static int	IX,IY,IZ,IX0,IY0,IZ0,IX1,IY1,IZ1,IX2,IY2,IZ2;

static void	Comparison(ix0,iy0,iz0,ix1,iy1,iz1,ix2,iy2,iz2)
int		ix0,iy0,iz0,ix1,iy1,iz1,ix2,iy2,iz2;
{
	int	ix=ix0+ix1+ix2,
		iy=iy0+iy1+iy2,
		iz=iz0+iz1+iz2;
	double	x=(double)ix,x2=x*x,
		y=(double)iy,y2=y*y,
		z=(double)iz,z2=z*z,
		l0_lc=(xx*x2+yy*y2+zz*z2+xy2*x*y+yz2*y*z+zx2*z*x)/(x2+y2+z2);
#ifdef	OBLATE
	if (l0_lc<L0_Lc) {
#else
	if (l0_lc>L0_Lc) {
#endif
	    IX=ix; IX0=ix0; IX1=ix1; IX2=ix2;
	    IY=iy; IY0=iy0; IY1=iy1; IY2=iy2;
	    IZ=iz; IZ0=iz0; IZ1=iz1; IZ2=iz2; L0_Lc=l0_lc;
	}
}

static void	Recursion(level,ix0,iy0,iz0,ix1,iy1,iz1,ix2,iy2,iz2)
int		level,ix0,iy0,iz0,ix1,iy1,iz1,ix2,iy2,iz2;
{
	int	ix3,iy3,iz3,ix4,iy4,iz4,ix5,iy5,iz5;

	if (level<=0)
	    Comparison(ix0,iy0,iz0,ix1,iy1,iz1,ix2,iy2,iz2);
	else {
	    Recursion(--level,ix3=ix0+ix1,iy3=iy0+iy1,iz3=iz0+iz1,
			      ix4=ix1+ix2,iy4=iy1+iy2,iz4=iz1+iz2,
			      ix5=ix2+ix0,iy5=iy2+iy0,iz5=iz2+iz0);
	    Recursion(  level,ix0*2,iy0*2,iz0*2,ix3,iy3,iz3,ix5,iy5,iz5);
	    Recursion(  level,ix1*2,iy1*2,iz1*2,ix4,iy4,iz4,ix3,iy3,iz3);
	    Recursion(  level,ix2*2,iy2*2,iz2*2,ix5,iy5,iz5,ix4,iy4,iz4);
	}
}

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif
#define M_PI2	(M_PI*2.0)

int	main(argc,argv)
int	argc;
char	**argv;
{
	char			**path;
	int			Nx,Ny,Nz,BPS,iy,iz,ix,level,
				ix0,iy0,iz0,ix1,iy1,iz1,ix2,iy2,iz2,
				ix3,iy3,iz3,ix4,iy4,iz4,ix5,iy5,iz5;
	double			Ux,Uy,Uz,z,y,x,
				tx,ty,l,cl,sl,csl,cl2,sl2,p,cp,sp,
				hh,vv,hv2,t,ct,st,cst,ct2,st2,L0_Lb,L0_La,tmp;
	Cell			**cell;
	unsigned		PV,pv;
	unsigned long long	n;

	static struct {
		unsigned long long	n;
		double			x,y,z,xx,yy,zz,xy,yz,zx;
	} sum[(size_t)(Cell)(~0)+1],*S;

	if (argc!=6)
	    Error("usage : si_of directory nameFile unitX unitY unitZ");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,&BPS);

	if ((Ux=atof(argv[3]))<=0.0 ||
	    (Uy=atof(argv[4]))<=0.0 ||
	    (Uz=atof(argv[5]))<=0.0) Error("bad unit length.");

	if ((cell=(Cell **)malloc(sizeof(Cell *)*(size_t)Ny))==NULL ||
	    (*cell=(Cell *)malloc(sizeof(Cell)*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for cell.");

	for (iy=1; iy<Ny; iy++) cell[iy]=cell[iy-1]+Nx;

	PV=(1<<BPS)-1;
	for (S=sum, pv=0; pv<=PV; pv++, S++) {
	    S->n=0; S->x=S->y=S->z=S->xx=S->yy=S->zz=S->xy=S->yz=S->zx=0.0;
	}
	for (z=Uz/2.0, iz=0; iz<Nz; iz++, z+=Uz) {
	    ReadSliceImage(path[iz],Nx,Ny,cell);

	    for (y=Uy/2.0, iy=0; iy<Ny; iy++, y+=Uy)
	    for (x=Ux/2.0, ix=0; ix<Nx; ix++, x+=Ux) {
		S=sum+cell[iy][ix]; ++S->n; S->x+=x; S->xx+=(x*x);
					    S->y+=y; S->yy+=(y*y); S->xy+=(x*y);
					    S->z+=z; S->zz+=(z*z); S->yz+=(y*z);
								   S->zx+=(z*x);
	    }
	}
	for (S=sum, pv=0; pv<=PV; pv++, S++)
	    if ((n=S->n)!=0) {

#define N	(double)n

		x=S->x/N; xx=S->xx+(Ux*Ux/12.0-x*x)*N;
		y=S->y/N; yy=S->yy+(Uy*Uy/12.0-y*y)*N; xy2=(S->xy-x*y*N)*2.0;
		z=S->z/N; zz=S->zz+(Uz*Uz/12.0-z*z)*N; yz2=(S->yz-y*z*N)*2.0;
						       zx2=(S->zx-z*x*N)*2.0;
#ifdef	OBLATE
		L0_Lc=xx+yy+zz+fabs(xy2)+fabs(yz2)+fabs(zx2);
#else
		L0_Lc=0.0;
#endif
		Recursion(RECURSION_LEVEL,0,0,1, 1, 0,0, 0, 1,0);
		Recursion(RECURSION_LEVEL,0,0,1, 0, 1,0,-1, 0,0);
		Recursion(RECURSION_LEVEL,0,0,1,-1, 0,0, 0,-1,0);
		Recursion(RECURSION_LEVEL,0,0,1, 0,-1,0, 1, 0,0);

		for (level=RECURSION_LEVEL; level<TOTAL_LEVEL; level++) {
ix=(IX*=2); ix0=IX0; ix1=IX1; ix2=IX2; ix3=IX0+IX1; ix4=IX1+IX2; ix5=IX2+IX0;
iy=(IY*=2); iy0=IY0; iy1=IY1; iy2=IY2; iy3=IY0+IY1; iy4=IY1+IY2; iy5=IY2+IY0;
iz=(IZ*=2); iz0=IZ0; iz1=IZ1; iz2=IZ2; iz3=IZ0+IZ1; iz4=IZ1+IZ2; iz5=IZ2+IZ0;

		    Comparison(ix0*2,iy0*2,iz0*2,ix3,iy3,iz3,ix5,iy5,iz5);
		    Comparison(ix1*2,iy1*2,iz1*2,ix4,iy4,iz4,ix3,iy3,iz3);
		    Comparison(ix2*2,iy2*2,iz2*2,ix5,iy5,iz5,ix4,iy4,iz4);

		    if (IX==ix && IY==iy && IZ==iz) {
			IX0=ix3; IX1=ix4; IX2=ix5;
			IY0=iy3; IY1=iy4; IY2=iy5;
			IZ0=iz3; IZ1=iz4; IZ2=iz5;
		    }
		}
		l=atan2(ty=(double)IY,tx=(double)IX);
		csl=(cl=cos(l))*(sl=sin(l)); cl2=cl*cl; sl2=sl*sl;
		p=atan2((double)IZ,sqrt(tx*tx+ty*ty)); cp=cos(p); sp=sin(p);

		hh=xx*sl2+yy*cl2-xy2*csl;
		vv=(xx*cl2+yy*sl2+xy2*csl)*sp*sp+zz*cp*cp-(zx2*cl+yz2*sl)*cp*sp;
		hv2=((xx-yy)*csl*2.0-xy2*(cl2-sl2))*sp-(zx2*sl-yz2*cl)*cp;

		t=fmod(atan2(hv2,hh-vv)+M_PI2,M_PI2)/2.0;
		cst=(ct=cos(t))*(st=sin(t)); ct2=ct*ct; st2=st*st;
		L0_Lb=hh*st2+vv*ct2-hv2*cst; L0_La=hh*ct2+vv*st2+hv2*cst;
#ifdef	OBLATE
		if (L0_Lb>L0_La) {
#else
		if (L0_Lb<L0_La) {
#endif
		    tmp=L0_Lb; L0_Lb=L0_La; L0_La=tmp;
		    t=fmod(t*2.0+M_PI,M_PI2)/2.0;
		}
		(void)printf("%u\t%" LL_FORMAT "u\t%lg\t%lg\t%lg\t"
			     "%lg\t%lg\t%lg\t%lg\t%lg\t%lg\n",
			     pv,n,x,y,z,l*180.0/M_PI,
					p*180.0/M_PI,
					t*180.0/M_PI,sqrt(L0_La*5.0/N),
						     sqrt(L0_Lb*5.0/N),
						     sqrt(L0_Lc*5.0/N));
	    }

	return 0;
}
