
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "csi.h"
#include "sif.h"

extern void	Error(),ReadSliceImage();

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif

#define LEN	1024
#define PV	(Cell)(~0)

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,line[LEN],form[LEN],path[LEN];
	int	Nx,Ny,Nz,pv,iy,tmp,ix1,ix2,iy1,iy2,iz1,iz2,iz,ix;
	Cell	bg,**cell,fg;
	double	Ux,Uy,Uz,x,y,z,l,A,B,C,cl,sl,p,cp,sp,t,ct,st,
		ax,bx,cx,ay,by,cy,az,bz,cz,axy,ayz,azx,bxy,byz,bzx,cxy,cyz,czx,
		X,Y,Z,Zzx,Zyz,xy,ZZzz,yy,xx,Yxy_Zzx,YYyy_YZyz_ZZzz;
	size_t	len;

	struct List {
		Cell		fg;
		int		ix1,ix2,iy1,iy2,iz1,iz2;
		double		x,y,z,xx,yy,zz,xy,yz,zx;
		struct List	*next;
	} *list,*next,**ptr=(&next);

	if (argc!=7 && argc!=9)
	Error("usage : si_op (orgDir nameFile)|(Nx Ny Nz PV) Ux Uy Uz newDir");

	if (argc==7)
	    paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);
	else {
	    if ((Nx=atoi(argv[1]))<=0 ||
		(Ny=atoi(argv[2]))<=0 ||
		(Nz=atoi(argv[3]))<=0) Error("bad image size.");

	    if ((pv=atoi(argv[4]))<0 || pv>PV)
		Error("bad pixel value for space filling.");

	    bg=(Cell)pv;
	}
	if ((cell=(Cell **)malloc(sizeof(Cell *)*(size_t)Ny))==NULL ||
	    (*cell=(Cell *)malloc(sizeof(Cell)*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatable memory for cell.");

	for (iy=1; iy<Ny; iy++) cell[iy]=cell[iy-1]+Nx;

	if ((Ux=atof(argv[argc-4]))<=0.0 ||
	    (Uy=atof(argv[argc-3]))<=0.0 ||
	    (Uz=atof(argv[argc-2]))<=0.0) Error("bad unit length.");

#define ERROR_LINE(msg)	(void)fputs(msg " :\n",stderr); Error(line)

	while (fgets(line,LEN,stdin)!=NULL) {
	    if (line[tmp=strlen(line)-1]!='\n') {
		ERROR_LINE("too long line");
	    }
	    line[tmp]='\0';

	    if (sscanf(line,"%d %lf %lf %lf %lf %lf %lf %lf %lf %lf",
			    &pv,&x,&y,&z,&l,&p,&t,&A,&B,&C)!=10) {
		ERROR_LINE("bad number of parameters");
	    }
	    if (pv<0 || pv>PV) {
		ERROR_LINE("bad pixel value for ovoid filling");
	    }
	    if (A<=0.0 || B<=0.0 || C<=0.0) {
		ERROR_LINE("bad radius");
	    }
	    x=x/Ux-0.5; l*=(M_PI/180.0); cl=cos(l); sl=sin(l); A*=A;
	    y=y/Uy-0.5; p*=(M_PI/180.0); cp=cos(p); sp=sin(p); B*=B;
	    z=z/Uz-0.5; t*=(M_PI/180.0); ct=cos(t); st=sin(t); C*=C;

	    ax=  sl*ct+cl*sp*st ; bx=(-sl*st+cl*sp*ct); cx=cl*cp;
	    ay=(-cl*ct+sl*sp*st); by=  cl*st+sl*sp*ct ; cy=sl*cp;
	    az=(         -cp*st); bz=(         -cp*ct); cz=   sp;
	    axy=ax*ay; ayz=ay*az; azx=az*ax; ax*=ax; ay*=ay; az*=az;
	    bxy=bx*by; byz=by*bz; bzx=bz*bx; bx*=bx; by*=by; bz*=bz;
	    cxy=cx*cy; cyz=cy*cz; czx=cz*cx; cx*=cx; cy*=cy; cz*=cz;

	    X=sqrt(A*ax+B*bx+C*cx)/Ux;
	    Y=sqrt(A*ay+B*by+C*cy)/Uy;
	    Z=sqrt(A*az+B*bz+C*cz)/Uz;

	    if ((ix1=(int)ceil(x-X))<Nx && (ix2=(int)floor(x+X)+1)>0 &&
		(iy1=(int)ceil(y-Y))<Ny && (iy2=(int)floor(y+Y)+1)>0 &&
		(iz1=(int)ceil(z-Z))<Nz && (iz2=(int)floor(z+Z)+1)>0) {
		if ((list=(struct List *)malloc(sizeof(struct List)))==NULL)
		    Error("no allocatable memory for list.");

		list->fg=(Cell)pv;
		list->ix1=(ix1<0)?0:ix1; list->ix2=(ix2>Nx)?Nx:ix2; list->x=x;
		list->iy1=(iy1<0)?0:iy1; list->iy2=(iy2>Ny)?Ny:iy2; list->y=y;
		list->iz1=(iz1<0)?0:iz1; list->iz2=(iz2>Nz)?Nz:iz2; list->z=z;
		list->xx=ax/A+bx/B+cx/C; list->xy=2.0*(axy/A+bxy/B+cxy/C);
		list->yy=ay/A+by/B+cy/C; list->yz=2.0*(ayz/A+byz/B+cyz/C);
		list->zz=az/A+bz/B+cz/C; list->zx=2.0*(azx/A+bzx/B+czx/C);
		*ptr=list; ptr=(&(list->next));
	    }
	}
	*ptr=NULL;

	if (argc==7)
	    len=strlen(argv[1])+1;
	else {
	    tmp=1; for (iz=10; iz<Nz; iz*=10) ++tmp;
	    (void)sprintf(form,"%s/%%0%dd.tif",argv[argc-1],tmp);
	}
	for (iz=0; iz<Nz; iz++) {
	    if (argc==7)
		ReadSliceImage(paths[iz],Nx,Ny,cell);
	    else
		for (iy=0; iy<Ny; iy++)
		for (ix=0; ix<Nx; ix++) cell[iy][ix]=bg;

	    for (list=next; list!=NULL; list=list->next)
		if (iz>=list->iz1 && iz<list->iz2) {
		    Z=((double)iz-list->z)*Uz;
		    Zzx=Z*list->zx; Zyz=Z*list->yz; xy=list->xy;
		    ZZzz=Z*Z*list->zz; yy=list->yy; xx=list->xx;
		    y=list->y; iy2=list->iy2; iy1=list->iy1;
		    x=list->x; ix2=list->ix2; ix1=list->ix1;
		    fg=list->fg;
		    for (Y=((double)(iy=iy1)-y)*Uy; iy<iy2; iy++, Y+=Uy) {
			Yxy_Zzx=Y*xy+Zzx; YYyy_YZyz_ZZzz=Y*Y*yy+Y*Zyz+ZZzz; 
			for (X=((double)(ix=ix1)-x)*Ux; ix<ix2; ix++, X+=Ux)
			    if (X*X*xx+X*Yxy_Zzx+YYyy_YZyz_ZZzz<=1.0) cell[iy][ix]=fg;
		    }
		}

	    pv=0;
	    for (iy=0; iy<Ny; iy++)
	    for (ix=0; ix<Nx; ix++) if (cell[iy][ix]>pv) pv=cell[iy][ix];

	    for (tmp=1; 1<<tmp<=pv; tmp<<=1) ;

	    if (argc==7)
		(void)sprintf(path,"%s/%s",argv[argc-1],paths[iz]+len);
	    else
		(void)sprintf(path,form,iz);

	    StoreImageFile(path,Nx,Ny,tmp,cell,NULL);
	}
	return 0;
}
