
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

extern void	Error();

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif

#define LEN	1024
#define PV	(Cell)(~0)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,pv,iy,ix,tmp,ix1,ix2,iy1,iy2;
	Cell	**cell;
	double	Ux,Uy,x,y,t,A,B,c,s,xy,xx,yy,X,Y,Yxy,YYyy;
	char	line[LEN];

	if (argc!=5 && argc!=7)
	    Error("usage : t_op orgTIFF|(Nx Ny PV) Ux Uy newTIFF");

	if (argc==5)
	    ReadImageFile(argv[1],&Nx,&Ny,NULL,&cell,NULL);
	else {
	    if ((Nx=atoi(argv[1]))<=0 ||
		(Ny=atoi(argv[2]))<=0) Error("bad image size.");

	    if ((pv=atoi(argv[3]))<0 || pv>PV)
		Error("bad pixel value for space filling.");

	    if ((cell=(Cell **)malloc(sizeof(Cell *)*(size_t)Ny))==NULL ||
		(*cell=(Cell *)malloc(sizeof(Cell)*(size_t)Ny
						  *(size_t)Nx))==NULL)
		Error("no allocatable memory for cell.");

	    for (iy=0; iy<Ny; iy++) {
		if (iy>0) cell[iy]=cell[iy-1]+Nx;

		for (ix=0; ix<Nx; ix++) cell[iy][ix]=(Cell)pv;
	    }
	}
	if ((Ux=atof(argv[argc-3]))<=0.0 ||
	    (Uy=atof(argv[argc-2]))<=0.0) Error("bad unit length.");

#define ERROR_LINE(msg)	(void)fputs(msg " :\n",stderr); Error(line)

	while (fgets(line,LEN,stdin)!=NULL) {
	    if (line[tmp=strlen(line)-1]!='\n') {
		ERROR_LINE("too long line");
	    }
	    line[tmp]='\0';

	    if (sscanf(line,"%d %lf %lf %lf %lf %lf",&pv,&x,&y,&t,&A,&B)!=6) {
		ERROR_LINE("bad number of parameters");
	    }
	    if (pv<0 || pv>PV) {
		ERROR_LINE("bad pixel value for oval filling");
	    }
	    if (A<=0.0 || B<=0.0) {
		ERROR_LINE("bad radius");
	    }
	    x=x/Ux-0.5; y=y/Uy-0.5; t*=(M_PI/180.0); A*=A; B*=B;
	    xy=2.0*(1.0/A-1.0/B)*(c=cos(t))*(s=sin(t)); c*=c; s*=s;
	    xx=c/A+s/B; yy=s/A+c/B; X=sqrt(A*c+B*s)/Ux; Y=sqrt(A*s+B*c)/Uy;

	    if ((ix1=(int)ceil(x-X))<Nx && (ix2=(int)floor(x+X)+1)>0 &&
		(iy1=(int)ceil(y-Y))<Ny && (iy2=(int)floor(y+Y)+1)>0) {
		if (ix1<0) ix1=0; if (ix2>Nx) ix2=Nx;
		if (iy1<0) iy1=0; if (iy2>Ny) iy2=Ny;

		for (Y=((double)(iy=iy1)-y)*Uy; iy<iy2; iy++, Y+=Uy) {
		    Yxy=Y*xy; YYyy=Y*Y*yy;
		    for (X=((double)(ix=ix1)-x)*Ux; ix<ix2; ix++, X+=Ux)
			if (X*X*xx+X*Yxy+YYyy<=1.0) cell[iy][ix]=(Cell)pv;
		}
	    }
	}
	pv=0;
	for (iy=0; iy<Ny; iy++)
	for (ix=0; ix<Nx; ix++) if (cell[iy][ix]>pv) pv=cell[iy][ix];

	for (tmp=1; 1<<tmp<=pv; tmp<<=1) ;
	StoreImageFile(argv[argc-1],Nx,Ny,tmp,cell,NULL);

	return 0;
}
